@@ -2,6 +2,7 @@ import curses#Depends on windows-curses on win32
 from curses.textpad import rectangle
 import enum
 from . import messagebox
+from . import filedialog
 import os
 from time import sleep
 import textwrap
@@ -138,10 +139,14 @@ def dynamic_search_and_select(stdscr,items:list[str],prompt:str,allowcancel=True
         utils.fill_line(stdscr,0,set_colour(BLUE,WHITE))
         stdscr.addstr(0,0,prompt,set_colour(BLUE,WHITE))      
 
-def cursesinput(stdscr,prompt: str,lines=1,maxlen=0,passwordchar:str=None,retremptylines=False,prefiltext="",bannedcharacters="") -> str:
+def cursesinput(stdscr,prompt: str,lines=1,maxlen=0,passwordchar:str=None,retremptylines=False,prefiltext="",bannedcharacters="",showhidecursor=True) -> str:
     """
     Get input from the user. Set maxlen to 0 for no maximum. Set passwordchar to None for no password entry. Retremptylines is if the program should return newlines even if the lines are empty. bannedcharacters is a comma-seperated list of banned words and characters
     """
+
+    if showhidecursor:
+        utils.showcursor()
+    
     ERROR = ""
     if passwordchar is not None:
         passworduse = True
@@ -193,6 +198,8 @@ def cursesinput(stdscr,prompt: str,lines=1,maxlen=0,passwordchar:str=None,retrem
         if ch == 10 or ch == 13 or ch == curses.KEY_ENTER :
             if lines == 1:
                 stdscr.erase()
+                if showhidecursor:
+                    utils.hidecursor()
                 return "\n".join(["".join(t) for t in text])
             if ln == lines - 1:
                 ERROR = " You have reached the bottom of the page. "
@@ -269,6 +276,35 @@ def cursesinput(stdscr,prompt: str,lines=1,maxlen=0,passwordchar:str=None,retrem
                     ln = 0
                     col = 0
                     stdscr.erase()
+                else:
+                    try:
+                        mappings = {
+                            b"SHF_PADENTER" : b"'",
+                            b"CTL_PADENTER" : b"\"",
+                            b"^@" : b"]"
+                        }
+                        chn = mappings[chn]
+                    except:
+                        continue
+                    #append
+                    if __calc_nbl_list(text) == maxlen and maxlen != 0:
+                        curses.beep()
+                        ERROR = f" You have reached the character limit ({maxlen}) "
+                    else:
+                        lll = False
+                        if bannedcharacters != "":
+                            for z in bannedcharacters.split(","):
+                                if z in textl or z == chn.decode():
+                                    lll = True
+                                    ERROR = " That is a banned character"
+                        if not lll:
+                            col += 1
+                            text[ln].insert(col-1,chn.decode())
+                            
+                            if col > mx-2:
+                                xoffset += 1
+                                stdscr.clear()
+
             else:
                 #append
                 if __calc_nbl_list(text) == maxlen and maxlen != 0:
@@ -473,13 +509,201 @@ def numericinput(stdscr,message="Please input a number",allowdecimals=False,allo
                 curses.beep()
                 continue
 
-def textview(stdscr,file=None,text=None,isagreement=False,requireyes=True,message="") -> bool:
+def _trail(ind,maxlen:int):
+    """Truncate and add ... to string if it is too long"""
+    if maxlen < 1:
+        return ""
+    ind = str(ind)
+    if len(ind) > maxlen:
+        return ind[0:maxlen-4]+"..."
+    else:
+        return ind
+
+def _dict_to_calctype(inputd) -> list:# Inputd must be dict or list. Fixed because of bug when running on py3.9
+    """Assemble list of types for dictedit. Not for use by standard users; internal only."""
+    if inputd is None:
+        return ["NULL!!"]
+    final = []
+    if type(inputd) == dict:
+        for ik in list(inputd.items()):
+            src = ik[0]
+            dst = ik[1]
+            if type(dst) == int or type(dst) == float:
+                obtype = "number"
+            elif type(dst) == str:
+                obtype = "string"
+            elif type(dst) == dict:
+                obtype = "folder"
+            elif type(dst) == list:
+                obtype = "list"
+            elif type(dst) == bool:
+                obtype = "bool"
+            else:
+                obtype = str(type(dst))
+            final.append(f"{src} ( {obtype} ) = {_trail(dst,os.get_terminal_size()[1]-20)}")
+    else:
+        oi = 0
+        for ik in list(inputd):
+            if type(ik) == int or type(ik) == float:
+                obtype = "number"
+            elif type(ik) == str:
+                obtype = "string"
+            elif type(ik) == dict:
+                obtype = "folder"
+            elif type(ik) == list:
+                obtype = "list"
+            elif type(ik) == bool:
+                obtype = "bool"
+            else:
+                obtype = str(type(ik))
+            final.append(f"Object {oi} ( {obtype} ) = {_trail(ik,os.get_terminal_size()[1]-30)}")
+            oi += 1
+    return final
+
+def _dictpath(inputd:dict,path:str):
+    """Select path from inputd where path is seperated by /"""
+    if path == "/":
+        return inputd
+    final = inputd
+    for axx in path.split("/"):
+        if axx == "":
+            continue
+        if type(final) == list:
+            final = final[int(axx)]
+        else:
+            final = final[axx]
+    return final
+
+
+def dictedit(stdscr,inputd:dict,name:str,isflat:bool=False) -> dict:
+    """
+    ## Edit Dictionaries Interactively
+
+    Dictedit provides a fancy screen set for the viewing and editing of dictionary structured.
+    Parameters:
+
+    inputd : the dictionary you want to edit
+    name: What you want the user to do / info message
+    isflat: If true, forbids the user from creating new {} and [], only flat values
+    """
+    path = "/"
+    while True:
+        mx,my = os.get_terminal_size()
+        path = path.replace("//","/")
+        options = ["Quit","Move up directory","ADD ITEM"]
+        currentedit = _dictpath(inputd,path)
+        options += _dict_to_calctype(currentedit)
+        e = coloured_option_menu(stdscr,options,f"{name} | Path: {path}",[
+            [
+                "quit",RED
+            ],["ADD",GREEN],
+            ["list",YELLOW],
+            ["folder",YELLOW],
+            ["string",CYAN],
+            ["number",MAGENTA]
+        ],preselected=2,footer="When editing a key: If you didn't mean to edit, press enter without doing anything")
+        if e == 0:
+            return inputd
+        elif e == 1:
+            if path != "/":
+                path = "/".join(path.split("/")[0:-1])
+            else:
+                continue
+        elif e == 2:
+            if type(_dictpath(inputd,path)) == dict:
+                keynamne = cursesinput(stdscr,"Please input the key name.")
+                ktype = coloured_option_menu(stdscr,["Cancel add","String","Number","Boolean (yes/no)","Empty list","Empty folder"],"What is the type of the key")
+                if ktype > 3 and isflat:
+                    messagebox.showerror(stdscr,["The selected type is not supported by this","file format"])
+                    continue
+                if ktype == 0:
+                    continue
+                elif ktype == 1:
+                    val = cursesinput(stdscr,"What is the value of this key?")
+                elif ktype == 2:
+                    val = numericinput(stdscr,"What is the value of this key?",True,True)
+                    if int(val) == val:
+                        val = int(val)
+                elif ktype == 3:
+                    val = messagebox.askyesno(stdscr,["New value for boolean?"])
+                elif ktype == 4:
+                    val = []
+                elif ktype == 5:
+                    val = {}
+                paths = [z for z in path.split("/") if z != ""]
+                mydata = inputd
+                for i in paths:
+                    if type(mydata) == list:
+                        mydata = mydata[int(i)]
+                    else:
+                        mydata = mydata[i]
+                mydata[keynamne] = val 
+            elif type(_dictpath(inputd,path)) == list:
+                #keynamne = crssinput(stdscr,"Please input the key name.")
+                ktype = coloured_option_menu(stdscr,["Cancel add","String","Number","Boolean (yes/no)","Empty list","Empty folder"],"What is the type of the key")
+                if ktype == 0:
+                    continue
+                elif ktype == 1:
+                    val = cursesinput(stdscr,"What is the value of this key?")
+                elif ktype == 2:
+                    val = numericinput(stdscr,"What is the value of this key?",True,True)
+                    if int(val) == val:
+                        val = int(val)
+                elif ktype == 3:
+                    val = messagebox.askyesno(stdscr,["New value for boolean?"])
+                elif ktype == 4:
+                    val = []
+                elif ktype == 5:
+                    val = {}
+                paths = [z for z in path.split("/") if z != ""]
+                mydata = inputd
+                for i in paths:
+                    if type(mydata) == list:
+                        mydata = mydata[int(i)]
+                    else:
+                        mydata = mydata[i]
+                mydata.append(val)
+        else:
+            newval = None
+            if type(currentedit) == dict:
+                path += "/" + list(currentedit.items())[e-3][0]
+            elif type(currentedit) == list:
+                path += "/" + str(e-3)
+            epath = path.split("/")[-1]
+            if type(_dictpath(inputd,path)) == str:
+                utils.showcursor()
+                newval = cursesinput(stdscr,f"Please input a new value for {path}",prefiltext=_dictpath(inputd,path))
+                utils.hidecursor()
+            elif type(_dictpath(inputd,path)) == int or type(_dictpath(inputd,path)) == float:
+                utils.showcursor()
+                newval = numericinput(stdscr,f"Please input a new value for {path}",True,True,prefillnumber=_dictpath(inputd,path))
+                utils.hidecursor()
+            elif type(_dictpath(inputd,path)) == bool:
+                if _dictpath(inputd,path):
+                    nv = messagebox.MessageBoxStates.YES
+                else:
+                    nv = messagebox.MessageBoxStates.NO
+                newval = messagebox.askyesno(stdscr,["New boolean value for",path],default=nv)
+            if type(_dictpath(inputd,path)) != dict and type(_dictpath(inputd,path)) != list:
+                path = "/".join(path.split("/")[0:-1]) 
+            if newval is not None:
+                paths = [z for z in path.split("/") if z != ""]
+                mydata = inputd
+                for i in paths:
+                    if type(mydata) == list:
+                        mydata = mydata[int(i)]
+                    else:
+                        mydata = mydata[i]
+                mydata[epath] = newval      
+
+def textview(stdscr,file=None,text=None,isagreement=False,requireyes=True,message="",allowprintout=True) -> bool:
     """
     ## View Text interactively
 
     This function is resize-friendly
     Set either file or text to not none to use mode. Set isagreement to true if you want this to be a license agreement
     Returns true if isagreement is false or if user agreed. Returns false if the user did not agree
+    set allowprintout to set if the user can print to a file or not
     """
     offset = 0
     if file is None and text is None:
@@ -490,28 +714,46 @@ def textview(stdscr,file=None,text=None,isagreement=False,requireyes=True,messag
     elif text is not None:
         text = text
 
+    displaymsg(stdscr,["Preparing text","Please wait..."],False)
     zltext = text.splitlines()
+    mx,my = os.get_terminal_size()
+    n = mx - 1
+    broken_text = []
+    for text in zltext:
+        if text.replace(" ","") == "":
+            broken_text += [""]
+        else:
+            broken_text += textwrap.wrap(text,n)
     while True:
         stdscr.clear()
         #stdscr.refresh()
         #Segment text
-        mx,my = os.get_terminal_size()
-        n = mx - 1
-        broken_text = []
-        for text in zltext:
-            if text.replace(" ","") == "":
-                broken_text += [""]
-            else:
-                broken_text += textwrap.wrap(text,n)
-        utils.fill_line(stdscr,0,set_colour(WHITE,BLACK))
-        stdscr.addstr(0,0,message[0:mx-8],set_colour(WHITE,BLACK))
+        if os.get_terminal_size()[0]-1 != n:
+            #Resegment text
+            mx,my = os.get_terminal_size()
+            n = mx - 1
+            broken_text = []
+            for text in zltext:
+                if text.replace(" ","") == "":
+                    broken_text += [""]
+                else:
+                    broken_text += textwrap.wrap(text,n)
+        utils.fill_line(stdscr,0,set_colour(BLUE,WHITE))
+        stdscr.addstr(0,0,message[0:mx-8],set_colour(BLUE,WHITE))
         prog = f"{offset}/{len(broken_text)}"
-        stdscr.addstr(0,mx-len(prog)-1,prog,set_colour(WHITE,BLACK))
-        utils.fill_line(stdscr,my-1,set_colour(WHITE,BLACK))
+        stdscr.addstr(0,mx-len(prog)-1,prog,set_colour(BLUE,WHITE))
+        utils.fill_line(stdscr,my-1,set_colour(BLUE,WHITE))
+        appenderstring = ""
         if isagreement:
-            stdscr.addstr(my-1,0,"A: Agree | D: Disagree",set_colour(WHITE,BLACK))
+            appenderstring += "A: Agree | D: Disagree"
         else:
-            stdscr.addstr(my-1,0,"Press enter to exit",set_colour(WHITE,BLACK))
+            appenderstring += "Press enter to exit"
+        if allowprintout:
+            appenderstring += " | S: Save to file"
+        #if isagreement:
+        #    stdscr.addstr(my-1,0,"A: Agree | D: Disagree",set_colour(BLUE,WHITE))
+        #else:
+        stdscr.addstr(my-1,0,appenderstring,set_colour(BLUE,WHITE))
         li = 1
         for line in broken_text[offset:offset+(my-2)]:
             try:
@@ -536,6 +778,28 @@ def textview(stdscr,file=None,text=None,isagreement=False,requireyes=True,messag
                 return False
             else:
                 messagebox.showwarning(stdscr,["You must agree to the license to proceed"])
+        elif ch == curses.KEY_END:
+            if len(broken_text) > my-2:
+                offset = len(broken_text) - my+2
+        elif ch == curses.KEY_PPAGE:
+            if offset > 10:
+                offset -= 10
+            else:
+                offset = 0
+
+        elif ch == curses.KEY_NPAGE:
+            offset += 10
+        elif ch == 115:
+            savefile = savefile_selector(stdscr,extension=".txt")
+            if savefile == "NONE":
+                continue
+            try:
+                with open(savefile,"w+") as f:
+                    f.write("\n".join(broken_text))
+            except Exception as e:
+                messagebox.showerror(stdscr,["Could not save.",f"Reason: {str(e)}"])
+            else:
+                messagebox.showinfo(stdscr,["Save successsful"])
 
 class PleaseWaitScreen:
     def __init__(self,stdscr,message=["Please Wait"]):
@@ -781,4 +1045,126 @@ def date_time_selector(stdscr,stype:DateTimeSelectorTypes,prompt:str,allow_past
                 else:
                     tx = copy.deepcopy(tmp)#If no errors, commit tmp back to dx
             except:
-                messagebox.showerror(stdscr,["That is an invalid value"])
\ No newline at end of file
+                messagebox.showerror(stdscr,["That is an invalid value"])
+
+def _is_a_valid_file_name(s:str) -> bool:
+    #Checks if a string is avalid file path
+    banned_characters = ["\"","*","<",">",":","|","?"]
+    for c in banned_characters:
+        if c in s:
+            return False
+    return True
+
+def savefile_selector(stdscr,defaultdirectory=os.getcwd(),extension="",enforce_extension=False) -> str:
+    """Select a file to save to. This will return the full path of the file. This will prompt the user to choose a folder and a filepath. If you provide an extension and set enforce_extension to true, it will automatically add the extension to the user's selection. If NONE is returned, that means the user sent CANCEL"""
+    directo = defaultdirectory
+    fileo = ""
+    while True:
+        p = coloured_option_menu(stdscr,["Finish",f"Directory: {directo}",f"File: {fileo}",f"{directo}{os.sep}{fileo}","Cancel"],colouring=[["finish",GREEN],["cancel",RED]])
+        if p == 0:
+            if fileo == "" or directo == "":
+                messagebox.showerror(stdscr,["Please fill in the form entirely."])
+                continue
+            final = directo + os.sep + fileo
+            if _is_a_valid_file_name(final):
+                return final
+            else:
+                messagebox.showerror(stdscr,["Not a valid path."])
+        elif p == 1:
+            directo = filedialog.openfolderdialog(stdscr,"Choose the directory for the file",directo,False)
+        elif p ==  2:
+            barefn = cursesinput(stdscr,"Write filename",prefiltext=fileo)
+            if enforce_extension:
+                if not barefn.endswith(extension):
+                    barefn += extension
+            fileo = barefn
+        elif p == 3:
+            return "NONE"
+
+def _friendly_positions(ins:int) -> str:
+    conv = str(ins)
+    if conv.endswith("1"):
+        conv += "st"
+    elif conv.endswith("2"):
+        conv += "nd"
+    elif conv.endswith("3"):
+        conv += "rd"
+    else:
+        conv += "th"
+    return conv
+
+def bargraph(stdscr,data:dict[str,int],message:str,unit="",sort=True,adjusty=False):
+    if len(data) == 0:
+        messagebox.showerror(stdscr,["No data"])
+        return
+    xoffset = 0
+    footerl = 1
+    selected = 0
+    if sort:
+        data = dict(sorted(data.items(), key=lambda x: x[1], reverse=True))
+    while True:
+        stdscr.clear()
+        utils.fill_line(stdscr,0,set_colour(BLUE,WHITE))
+        stdscr.addstr(0,0,message+" | Press Q to quit | Press S to export",set_colour(BLUE,WHITE))
+        maxval = max(list(data.values()))
+        if adjusty:
+            minval = min(list(data.values()))
+        else:
+            minval = 0
+        if maxval-minval == 0:
+            minval = maxval-1
+        my,mx = stdscr.getmaxyx()
+        my -= 1 #Strange bug
+        graphspace = my - 1 - footerl #Remove head and foot
+        ti = 4
+        #Add numbers]
+        for i in range(graphspace):
+            stdscr.addstr(my-(i+footerl),0,str(round(i/graphspace*(maxval-minval))+minval))
+        for pset in list(data.items())[xoffset:xoffset+mx-5]:
+            ti += 1
+            pval = pset[1]
+            trueoffset = round(((pval-minval)/(maxval-minval))*graphspace)+1
+            
+            for i in range(trueoffset):             
+                if ti-5+xoffset == selected:
+                    try:
+                        stdscr.addstr(my-(i+footerl),ti,"█",set_colour(CYAN,CYAN))
+                    except:
+                        pass
+                else:
+                    try:
+                        stdscr.addstr(my-(i+footerl),ti,"█")
+                    except:
+                        pass
+        dnsel = list(data.keys())[selected]
+        stdscr.addstr(my,0,f"{dnsel} ({list(data.values())[selected]} {unit}) - {_friendly_positions(sorted(list(data.values()),reverse=True).index(list(data.values())[selected])+1)} place")
+        stdscr.refresh()
+        ch = stdscr.getch()
+        if ch == curses.KEY_LEFT:
+            if selected > 0:
+                selected -= 1
+            if selected-xoffset <= 0 and xoffset > 0:
+                xoffset -= 1
+        elif ch == curses.KEY_RIGHT:
+            if selected < len(data)-1:
+                selected += 1
+            if selected-xoffset > mx-7:
+                xoffset += 1
+        elif ch == curses.KEY_SRIGHT:
+            xoffset += 1
+        elif ch == curses.KEY_SLEFT:
+            if xoffset > 0:
+                xoffset -= 1
+        elif ch == 115:
+            filepath = savefile_selector(stdscr,extension=".csv",enforce_extension=True)#You had better put a CSV file
+            if filepath == "NONE":
+                continue
+            with open(filepath,"w+") as f:
+                f.write(f"Key,{unit},{message}")
+                f.write("\n")
+                for pair in list(data.items()):
+                    f.write(f"{pair[0]},{pair[1]}")
+                    f.write("\n")
+            messagebox.showinfo(stdscr,["Save successful"])
+        elif ch == 113:
+            return
\ No newline at end of file
